# -------------------------------------------------------------------------- #
# backward probability
#
# b(i, t1, t2) = P(O_{t2+1}, ..., O_T | S_i at t1:t2, lambda)
#
# -------------------------------------------------------------------------- #

"backward" <- 
function(x, p, a, du, D, step, x.end){
    
    check.input(x, p, a, du, D)
    
    ## number of observiations
    n = length(x)
    
    ## number of states
    m = nrow(a)
        
    ## make sure p start from a value < step
    p = p - floor(min(p)/step)*step

    ## how many steps altogether
    Z = ceiling(max(p)/step)
    
    ## step indicator, every 'step' bp is one step
    stp = ceiling(p/step)
    
    ## check that for the input data, every step have at least 2 observations
    len = tapply(p, stp, length)
    if(any(len<2)){
        stop("there is at least one step with less than 2 observations\n")
    }
    
    ## indicator of end of every step
    dif = diff(stp)
    stp.end = c(which(dif>0), length(stp))

    ## end of each step
    x1.end = x.end[[1]]
    x2.end = x.end[[2]]
    x3.end = x.end[[3]]
    x4.end = x.end[[4]]

    ## log probability
    log.a  = log(a)
    log.du = log(du)

    ## backward probability
    b1 = b2 = b3 = b4 = matrix(-1/0, nrow=Z, ncol=Z)

    ## initialization
    b1[,Z] = rep(0, Z) ## log(rep(1, m))
    b2[,Z] = rep(0, Z)
    b3[,Z] = rep(0, Z)
    b4[,Z] = rep(0, Z)

    for(t2 in (Z-1):1){
        if(t2 %% 20 == 0){
            cat("t2 =", t2, "\n")
        }
        ## (tn1, tn2) are (t1, t2) in next state
        tn1  = t2+1 
        tn2s = tn1:min(t2+D,Z) 
        for(t1 in max(1,t2-D+1):t2){
            ## emission probability
            ev13 = ev24 = ev32 = ev34 = ev41 = numeric(length(tn2s))
            c=0
            for(tn2 in tn2s){
                c = c + 1
                kp = (stp.end[t2]+1):stp.end[tn2]
                x.kp = x[kp]
                p.kp = p[kp]
                
                state = c(0, 0, 1, 0)
                ev = emiss(x.kp, p.kp, x1.end[t1,t2], t2*step, state)
                ev13[c] = ev[3]

                state = c(0, 0, 0, 1)
                ev = emiss(x.kp, p.kp, x2.end[t1,t2], t2*step, state)
                ev24[c] = ev[4]

                state = c(0, 1, 0, 1)
                ev = emiss(x.kp, p.kp, x3.end[t1,t2], t2*step, state)
                ev32[c] = ev[2]
                ev34[c] = ev[4]
                
                state = c(1, 0, 0, 0)
                ev = emiss(x.kp, p.kp, x4.end[t1,t2], t2*step, state)
                ev41[c] = ev[1]
            }
            
            ds = tn2s - t2
            tmp1 = b1[tn1,tn2s]+log.du[ds,1]
            tmp2 = b2[tn1,tn2s]+log.du[ds,2]
            tmp3 = b3[tn1,tn2s]+log.du[ds,3]
            tmp4 = b4[tn1,tn2s]+log.du[ds,4]

            ## skip those terms where a[i,j] = 0
            b1[t1,t2] = log.a[1,3] + logsumexp(tmp3 + ev13)
            b2[t1,t2] = log.a[2,4] + logsumexp(tmp4 + ev24)
            v = numeric(2)
            v[1] = log.a[3,2] + logsumexp(tmp2 + ev32)
            v[2] = log.a[3,4] + logsumexp(tmp4 + ev34)
            b3[t1,t2] = logsumexp(v)
            b4[t1,t2] = log.a[4,1] + logsumexp(tmp1 + ev41)
        }
    }
    
    list(b1, b2, b3, b4)
}
